import importlib
import inspect
import os
import re

import pytest

from .config import get_store_reader

get_env = get_store_reader()


def get_marker(request):
    def with_value(value):
        marker = request.node.get_closest_marker(value)
        if marker is None:
            return None
        return marker.args[0]
    return with_value


def get_driver_params(env_var, default_values):
    requested_drivers = get_env(env_var)
    if requested_drivers is not None:
        return requested_drivers.split(',')
    return default_values


def get_driver_name(request):
    drivers_to_skip = get_marker(request)('skip_drivers') or []
    for driver_to_skip in drivers_to_skip:
        name = driver_to_skip.get('name', None)
        reason = driver_to_skip.get('reason', 'skipped for this driver: {}'.format(driver_to_skip))
        if name == request.param:
            pytest.skip(reason)
            break
    return request.param


def get_driver(name):
    def with_functionality(functionality_name):
        functionality = f"{name}_{functionality_name}"
        return importlib.import_module(functionality).execute
    return with_functionality



from datetime import datetime

from requests.auth import AuthBase

from verifit.driver import get_driver

from .cache import cache_set, cache_get
from .config import get_store_reader, get_store_writer
from .date_and_time import date_diff_in_minutes
from .iam_token import decode_token, get_token_expiration_date

get_env = get_store_reader()
set_env = get_store_writer()

KEY_ACCESS_TOKEN = 'access_token'
KEY_EXPIRY_DATE = 'expiry_date'
KEY_LOGIN_DATA = 'login_data'
KEY_USERNAME = 'username'
KEY_PASSWORD = 'password'


def get_access_token(login_data):
    return login_data.get(KEY_ACCESS_TOKEN, None)


def get_access_token_from_env():
    login_data = get_env(KEY_LOGIN_DATA)
    print(f"Using cached login data <{login_data}>")
    return get_access_token(login_data)


def get_expiry_date(login_data):
    return login_data.get(KEY_EXPIRY_DATE, None)


def get_expiry_date_from_env():
    return get_expiry_date(get_env(KEY_LOGIN_DATA))


def create_login_data():
    login_data = {}

    def with_access_token(access_token):
        login_data[KEY_ACCESS_TOKEN] = access_token

        def with_expiry_date(expiry_date):
            login_data[KEY_EXPIRY_DATE] = expiry_date
            return login_data

        return with_expiry_date

    return with_access_token


def get_login_user_name(user):
    return user.get(KEY_USERNAME, None)


def get_login_user_password(user):
    return user.get(KEY_PASSWORD, None)


def create_login_user():
    login_user = {}

    def with_username(username):
        login_user[KEY_USERNAME] = username

        def with_password(password):
            login_user[KEY_PASSWORD] = password
            return login_user

        return with_password

    return with_username


def login(user):
    def with_driver(driver):
        access_token = driver(user)
        decoded_token = decode_token(access_token)
        token_expiry_date = get_token_expiration_date(decoded_token)
        cache_set(get_login_user_name(user), {
            'accessToken': access_token,
            'expiryDate': token_expiry_date.isoformat()
        })
        login_data = create_login_data()(access_token)(token_expiry_date)
        set_env(KEY_LOGIN_DATA)(login_data)
        return login_data
    return with_driver


def login_main_user(driver_name):
    user = get_driver(driver_name)('user')('MAIN')
    return login(user)(get_driver(driver_name)('login'))


def login_main_user_from_cache(driver_name):
    user = get_driver(driver_name)('user')('MAIN')
    return login_from_cache(user)(get_driver(driver_name)('login'))


def get_login_values_from_cache(username):
    user_cached_data = cache_get(username)
    if user_cached_data is None:
        print(f"Cache miss key <{username}>, no cached data")
        return None
    token_expiry_iso_string = user_cached_data.get('expiryDate', None)
    if token_expiry_iso_string is None:
        print(f"Cache miss key <{username}>, no expiration info")
        return None
    token_expiry_date = datetime.fromisoformat(token_expiry_iso_string)
    minutes = date_diff_in_minutes(token_expiry_date, datetime.now())
    if minutes < 10:
        print(f"Cache miss key <{username}>, almost expired")
        return None
    print(f"Cache hit key <{username}>")
    return create_login_data()(user_cached_data.get('accessToken', None))(token_expiry_date)


def login_from_cache(user):
    def with_driver(driver):
        login_data = get_login_values_from_cache(get_login_user_name(user))
        if login_data is None:
            login_data = login(user)(driver)
        print(f"Caching login data <{login_data}>")
        set_env(KEY_LOGIN_DATA)(login_data)
        return login_data
    return with_driver


def get_bearer_authorization_header_value():
    return f"Bearer {get_access_token_from_env()}"


def get_bearer_auth_base():
    class BearerAuth(AuthBase):
        def __init__(self, token):
            self.token = token

        def __call__(self, r):
            r.headers["authorization"] = "Bearer " + self.token
            return r

    return BearerAuth(get_access_token_from_env())



import random
import string


def generate_random_digits_string(count=None):
    if count is None:
        count = random.randint(5, 10)
    result = ""
    for i in range(0, count):
        nr = random.randint(0, 9)
        result = f"{result}{nr}"
    return result


def generate_random_letters_string(count=None):
    if count is None:
        count = random.randint(5, 10)
    alphabet = string.ascii_lowercase + string.digits
    return ''.join(random.choice(alphabet) for _ in range(count))



import asyncio
import json

import websockets

from .config import get_store_reader

get_env = get_store_reader()


def ws_send_and_receive(a_dict):
    log_type = "[WebSockets]"
    wait_ms = int(get_env('WEBSOCKETS_WAIT_MS'))

    def with_server(server):
        def with_ignore_list(ignore_list):
            async def receive_async():
                async with websockets.connect(server) as websocket:
                    received = False
                    while not received:
                        data = await websocket.recv()
                        if data in ignore_list:
                            print(f"Ignoring received data <{data}>")
                        else:
                            received = True
                return data

            async def receive_async_with_timeout():
                seconds = wait_ms / 1000
                data = await asyncio.wait_for(
                    receive_async(),
                    seconds
                )
                print(f"{log_type} Received <{data}>")
                return data

            def start_receiving():
                print(f"{log_type} Waiting {wait_ms} ms for packets from {server}")
                return asyncio.ensure_future(receive_async_with_timeout())

            def end_receiving(a_handle):
                data = asyncio.get_event_loop().run_until_complete(a_handle)
                print(f"{log_type} Finished waiting for packets from {server}, received {data}")
                return json.loads(data)

            async def send_async(data):
                async with websockets.connect(server) as websocket:
                    await websocket.send(data)
                    print(f"Sent data <{data}>")

            def send_sync(data):
                asyncio.get_event_loop().run_until_complete(send_async(data))

            handle = start_receiving()
            to_send = json.dumps(a_dict)
            send_sync(to_send)
            return end_receiving(handle)
        return with_ignore_list
    return with_server




def date_diff_in_minutes(date1, date2):
    difference = date1 - date2
    minutes = difference.seconds // 60
    if difference.days < 0:
        minutes = minutes + difference.days * 24 * 60
    return minutes
